from .dom import TreeDom
from . import _rustlib
import typing


class Parser:
    __slots__ = ("__raw", "__state")

    def __init__(self, options: typing.Union[_rustlib.HtmlOptions, _rustlib.XmlOptions]):
        """
        An HTML/XML parser, ready to receive unicode input.

        This is very easy to use and allows you to stream input using `.process()` method; By this way
        you are don't worry about memory usages of huge inputs.

        for `options`, If your input is a HTML document, pass a `HtmlOptions`;
        If your input is a XML document, pass `XmlOptions`.
        """
        self.__raw = _rustlib.Parser(options)

        # 0 - processing
        # 1 - finished
        # 2 - converted
        self.__state = 0

    def process(self, content: typing.Union[str, bytes]) -> None:
        """
        Processes an input.

        `content` must be `str` or `bytes`.

        Raises `RuntimeError` if `.finish()` method is called.
        """
        self.__raw.process(content)

    def finish(self) -> None:
        """
        Finishes the parser and marks it as finished.

        Raises `RuntimeError` if is already finished.
        """
        self.__raw.finish()
        self.__state = 1

    def into_dom(self) -> TreeDom:
        """Converts the self into `TreeDom`. after calling this method, this object is unusable and you cannot use it."""
        dom = TreeDom(raw=self.__raw.into_dom())
        self.__state = 2
        return dom

    def errors(self) -> typing.List[str]:
        """
        Returns the errors which are detected while parsing.
        """
        return self.__raw.errors()

    @property
    def quirks_mode(self) -> int:
        """Returns the quirks mode (always is QUIRKS_MODE_OFF for XML)."""
        return self.__raw.quirks_mode()

    @property
    def lineno(self) -> int:
        """Returns the line count of the parsed content (always is `1` for XML)."""
        return self.__raw.lineno()

    @property
    def is_finished(self) -> bool:
        """Returns `True` if the parser is marked as finished"""
        return self.__state != 0

    @property
    def is_converted(self) -> bool:
        """Returns `True` if the parser is converted to `TreeDom` and now is unusable."""
        return self.__state == 2

    def __repr__(self) -> str:
        return repr(self.__raw)


def parse(
    content: typing.Union[str, bytes],
    options: typing.Union[_rustlib.HtmlOptions, _rustlib.XmlOptions],
) -> TreeDom:
    """
    Parses your HTML (or XML depends on `options` type) content and returns the parsed document tree.
    """
    parser = Parser(options)
    parser.process(content)
    parser.finish()
    return parser.into_dom()


def parse_file(
    path: typing.Union[str, typing.TextIO, typing.BinaryIO],
    options: typing.Union[_rustlib.HtmlOptions, _rustlib.XmlOptions],
    *,
    chunk_size: int = 10240,
) -> TreeDom:
    """
    Parses your HTML (or XML depends on `options` type) file and returns the parsed document tree.
    """
    from pathlib import Path

    close = False

    if isinstance(path, Path):
        path = str(path)

    if isinstance(path, str):
        path = open(path, "rb")
        close = True

    try:
        parser = Parser(options)

        while True:
            content = path.read(chunk_size)
            if not content:
                break

            parser.process(content)

        parser.finish()
        return parser.into_dom()
    finally:
        if close:
            path.close()



from . import _rustlib
import typing


if typing.TYPE_CHECKING:  # pragma: no cover
    from . import dom


class _IteratorMetaClass:
    """Bridge _rustlib iterators to Python"""

    _BASECLASS: typing.Callable[["dom.BaseNode"], typing.Iterable]

    __slots__ = ("_raw",)

    def __init__(self, value: "dom.BaseNode"):
        self._raw = iter(self._BASECLASS(value._raw))

    def __iter__(self):
        """Returns `iter(self)`"""
        return self

    def __next__(self) -> "dom.BaseNode":
        """Returns `next(self)`"""
        from .dom import BaseNode

        return BaseNode._wrap(next(self._raw))


class Ancestors(_IteratorMetaClass):
    """Iterates over ancestors (parents)."""

    _BASECLASS = _rustlib.iter.Ancestors


class PrevSiblings(_IteratorMetaClass):
    """Iterates over previous siblings."""

    _BASECLASS = _rustlib.iter.PrevSiblings


class NextSiblings(_IteratorMetaClass):
    """Iterates over next siblings."""

    _BASECLASS = _rustlib.iter.NextSiblings


class FirstChildren(_IteratorMetaClass):
    """Iterates over first children."""

    _BASECLASS = _rustlib.iter.FirstChildren


class LastChildren(_IteratorMetaClass):
    """Iterates over last children."""

    _BASECLASS = _rustlib.iter.LastChildren


class Children(_IteratorMetaClass):
    """Iterates children of a node."""

    _BASECLASS = _rustlib.iter.Children


class EdgeTraverse:
    """Open or close edge of a node (The returning type of `Traverse`)."""

    __slots__ = ("node", "closed")

    def __init__(self, node: "dom.BaseNode", closed: bool) -> None:
        self.node = node
        self.closed = closed

    def __repr__(self):
        if self.closed:
            return f"EdgeTraverse[closed]({self.node})"

        return f"EdgeTraverse[opened]({self.node})"


class Traverse(_IteratorMetaClass):
    """Iterator which traverses a tree."""

    _BASECLASS = _rustlib.iter.Traverse

    def __next__(self) -> EdgeTraverse:
        from .dom import BaseNode

        rn, closed = next(self._raw)
        return EdgeTraverse(BaseNode._wrap(rn), closed)


class Descendants(_IteratorMetaClass):
    """Iterates over a node and its descendants."""

    _BASECLASS = _rustlib.iter.Descendants


class Select:
    """An iterator that uses CSS selectors to match and find nodes."""

    __slots__ = ("__raw", "__limit", "__offset")

    def __init__(
        self, value: "dom.BaseNode", expr: str, *, limit: int = 0, offset: int = 0
    ) -> None:
        self.__raw = iter(_rustlib.Select(value._raw, expr))

        self.__limit = limit or -1
        self.__offset = offset - 1

    def __iter__(self):
        return self

    def __next__(self) -> "dom.Element":
        from .dom import Element

        if self.__limit == 0:
            raise StopIteration

        while self.__offset > 0:
            next(self.__raw)
            self.__offset -= 1

        node = Element(next(self.__raw))
        self.__limit -= 1

        return node



import markupever._rustlib as rl
import pytest


def test_qualname():
    q = rl.QualName("div")
    assert q.local == "div"
    assert q.ns == ""
    assert q.prefix is None

    q = rl.QualName("div", "html")
    assert q.local == "div"
    assert q.ns == "http://www.w3.org/1999/xhtml"
    assert q.prefix is None

    q = rl.QualName("div", "https://namespace1.org", prefix="ns1")
    assert q.local == "div"
    assert q.ns == "https://namespace1.org"
    assert q.prefix == "ns1"

    assert hash(q) == hash(q.copy())

    q1 = rl.QualName("a")
    q2 = rl.QualName("b")

    assert q1 < q2
    assert q1 != q2

    assert q1 != 1
    assert q1 == "a"
    assert q1 != "b"

    with pytest.raises(TypeError):
        q1 >= 1

    repr(q1)


def test_options():
    repr(rl.HtmlOptions())
    repr(rl.XmlOptions())


def test_parser_generators():
    parser = rl.Parser(rl.HtmlOptions())
    parser.process(b"<html><p>Ali</p></html>")
    parser.finish()

    repr(parser)

    parser = rl.Parser(rl.HtmlOptions())
    parser.process("<html><p>Ali</p></html>")

    repr(parser)

    with pytest.raises(TypeError):
        parser.process(1)

    with pytest.raises(RuntimeError):
        parser.into_dom()

    parser.finish()

    with pytest.raises(RuntimeError):
        parser.process("")

    with pytest.raises(RuntimeError):
        parser.finish()

    parser.into_dom()
    with pytest.raises(RuntimeError):
        parser.into_dom()

    parser = rl.Parser(rl.XmlOptions())
    for c in ("<html>", b"Ali", b"</html>"):
        parser.process(c)
    parser.finish()

    assert isinstance(parser.errors(), list)
    assert parser.lineno() == 1
    assert parser.quirks_mode() == 2

    parser = rl.Parser(rl.HtmlOptions(full_document=False))
    for c in (b"<html><p>Ali</p>", "\n", "</html>"):
        parser.process(c)
    parser.finish()

    assert parser.lineno() == 2

    _ = parser.into_dom()

    repr(parser)

    with pytest.raises(RuntimeError):
        parser.errors()


def test_document():
    dom = rl.TreeDom()

    with pytest.raises(NotImplementedError):
        rl.Document(dom)

    assert isinstance(dom.root(), rl.Document)
    assert dom.root() == dom.root()

    repr(dom)


def test_doctype():
    dom = rl.TreeDom()
    doctype = rl.Doctype(dom, "html", "", system_id="hello")

    with pytest.raises(TypeError):
        rl.Doctype(doctype, "xml")

    with pytest.raises(TypeError):
        rl.Doctype(1, "xml")

    assert doctype.parent() is None  # make sure it is orphan

    assert doctype.name == "html"
    assert doctype.system_id == "hello"
    assert doctype.public_id == ""

    doctype.name = "xml"
    doctype.public_id = "test"

    assert doctype.name == "xml"
    assert doctype.system_id == "hello"
    assert doctype.public_id == "test"

    repr(doctype)


def test_comment():
    dom = rl.TreeDom()
    x = rl.Comment(dom, "test")

    with pytest.raises(TypeError):
        rl.Comment(x, "xml")

    with pytest.raises(TypeError):
        rl.Comment("", "xml")

    assert x.parent() is None  # make sure it is orphan

    assert x.content == "test"
    x.content = "I am comment"
    assert x.content == "I am comment"

    repr(x)


def test_text():
    dom = rl.TreeDom()
    x = rl.Text(dom, "test")

    with pytest.raises(TypeError):
        rl.Text(x, "xml")

    with pytest.raises(TypeError):
        rl.Text("", "xml")

    assert x.parent() is None  # make sure it is orphan

    assert x.content == "test"
    x.content = "I am text"
    assert x.content == "I am text"

    repr(x)


def test_element():
    dom = rl.TreeDom()
    x = rl.Element(dom, "body", [], False, False)

    with pytest.raises(TypeError):
        rl.Element(x, "d", [], False, False)

    with pytest.raises(TypeError):
        rl.Element(b"", "d", [], False, False)

    assert x.parent() is None  # make sure it is orphan

    assert x.name == rl.QualName("body", "", None)
    assert isinstance(x.attrs, rl.AttrsList)
    assert x.template is False
    assert x.mathml_annotation_xml_integration_point is False

    x = rl.Element(dom, rl.QualName("div", "html", "ns"), [], False, True)

    assert x.name == rl.QualName("div", "html", "ns")
    assert isinstance(x.attrs, rl.AttrsList)
    assert x.template is False
    assert x.mathml_annotation_xml_integration_point is True

    with pytest.raises(TypeError):
        rl.Element(dom, rl.QualName("div", "html", "ns"), {}, False, True)

    rl.Element(dom, rl.QualName("div", "html", "ns"), [("a", "b")], False, True)
    rl.Element(dom, rl.QualName("div", "html", "ns"), [("a", "b"), ("c", "d")], False, True)
    rl.Element(
        dom, rl.QualName("div", "html", "ns"), [(rl.QualName("a"), "b"), ("c", "d")], False, True
    )
    rl.Element(
        dom, rl.QualName("div", "html", "ns"), ((rl.QualName("a"), "b"), ("c", "d")), False, True
    )

    with pytest.raises(TypeError):
        rl.Element(
            dom,
            rl.QualName("div", "html", "ns"),
            [(rl.QualName("a"), rl.QualName("b"))],
            False,
            True,
        )

    with pytest.raises(TypeError):
        rl.Element(dom, rl.QualName("div", "html", "ns"), [rl.QualName("a")], False, False)

    x = rl.Element(dom, rl.QualName("div", "html", "ns"), [], False, True)

    assert x.name == rl.QualName("div", "html", "ns")
    assert isinstance(x.attrs, rl.AttrsList)
    assert x.template is False
    assert x.mathml_annotation_xml_integration_point is True

    x.template = True
    x.mathml_annotation_xml_integration_point = False
    x.name = rl.QualName("html")
    x.attrs = []
    x.attrs = [(rl.QualName("a"), "b"), ("c", "d")]

    assert x.name == rl.QualName("html")
    assert isinstance(x.attrs, rl.AttrsList)
    assert x.template is True
    assert x.mathml_annotation_xml_integration_point is False

    x.name = "template"

    assert x.name == rl.QualName("template")

    repr(x)


def _get_attr(attrs: rl.AttrsList, name):
    for index, v in enumerate(attrs.items()):
        if v[0] == name:
            return index, v[1]

    return None


def test_element_attrs():
    dom = rl.TreeDom()
    x = rl.Element(dom, "body", [("class", "flex"), ("id", "main")], False, False)
    x = rl.Element(
        dom, "body", [(rl.QualName("class"), "flex"), (rl.QualName("id"), "main")], False, False
    )

    with pytest.raises(TypeError):
        rl.Element(dom, "wolf", (1, ""), False, False)

    with pytest.raises(TypeError):
        rl.Element(dom, "temple", {}, False, False)

    with pytest.raises(TypeError):
        rl.Element(dom, "hello", 0, False, False)

    assert len(x.attrs) == 2
    x.attrs.clear()
    assert len(x.attrs) == 0

    x.attrs.push("id", "panel")
    x.attrs.push(rl.QualName("class", "html"), "flex")

    assert _get_attr(x.attrs, "id") == (0, "panel")
    assert _get_attr(x.attrs, rl.QualName("id")) == (0, "panel")

    assert _get_attr(x.attrs, "class") == (1, "flex")
    assert _get_attr(x.attrs, rl.QualName("class", "html")) == (1, "flex")
    assert _get_attr(x.attrs, rl.QualName("class")) is None
    assert _get_attr(x.attrs, rl.QualName("class", "xml")) is None

    index, _ = _get_attr(x.attrs, "id")
    x.attrs.update_value(index, "x")

    assert _get_attr(x.attrs, "id") == (0, "x")

    with pytest.raises(IndexError):
        x.attrs.update_value(10, "x")

    x.attrs.remove(0)
    assert len(x.attrs) == 1
    assert _get_attr(x.attrs, "id") is None

    x.attrs.swap_remove(0)
    assert len(x.attrs) == 0
    assert _get_attr(x.attrs, "class") is None

    x = rl.Element(
        dom, "body", [("class", "flex"), ("class", "main"), ("class", "main")], False, False
    )

    assert len(x.attrs) == 3
    x.attrs.dedup()
    assert len(x.attrs) == 2

    assert _get_attr(x.attrs, "class") == (0, "flex")
    x.attrs.reverse()
    assert _get_attr(x.attrs, "class") == (0, "main")

    repr(x.attrs)


def test_pi():
    dom = rl.TreeDom()
    x = rl.ProcessingInstruction(dom, "d", target="t")

    with pytest.raises(TypeError):
        rl.ProcessingInstruction(x, "d", "t")

    with pytest.raises(TypeError):
        rl.ProcessingInstruction("", "d", "t")

    assert x.parent() is None  # make sure it is orphan

    assert x.data == "d"
    assert x.target == "t"

    x.data = "I am data"

    assert x.data == "I am data"
    assert x.target == "t"

    repr(x)


def test_append_prepend():
    # this test is enough because these methods uses ego_tree crate of Rust which is completely tested
    dom = rl.TreeDom()
    doctype = rl.Doctype(dom, "html", "", system_id="hello")

    dom.append(dom.root(), doctype)

    assert doctype.parent() == dom.root()
    assert dom.root().first_child() == doctype

    x = rl.Element(
        dom, rl.QualName("head", "mynamespace"), [("class", "flex"), ("id", "main")], False, False
    )

    assert dom.namespaces() == {}

    dom.prepend(dom.root(), x)

    assert dom.namespaces() == {"": "mynamespace"}

    assert x.parent() == dom.root()
    assert dom.root().first_child() == x

    y = rl.Element(
        dom,
        rl.QualName("head", "namespace2", "ns"),
        [("class", "flex"), ("id", "main")],
        False,
        False,
    )

    assert dom.namespaces() == {"": "mynamespace"}
    dom.insert_before(x, y)

    assert dom.namespaces() == {"": "mynamespace", "ns": "namespace2"}

    y = rl.Element(
        dom, rl.QualName("head", "namespace3"), [("class", "flex"), ("id", "main")], False, False
    )
    dom.insert_after(x, y)

    assert dom.namespaces() == {"": "mynamespace", "ns": "namespace2"}


def test_iterator():
    dom = rl.TreeDom()

    testcase = [
        rl.Doctype(dom, "one", "", ""),
        rl.Element(dom, "two", [], False, False),
        rl.Text(dom, "three"),
        rl.Comment(dom, "four"),
        rl.Element(dom, "five", [("class", "flex flex-row")], False, False),
        rl.Element(dom, rl.QualName("html", "html", None), [], False, False),
        rl.ProcessingInstruction(dom, "1", "2"),
    ]

    for index, node in enumerate(rl.iter.Iterator(dom)):
        if isinstance(node, rl.Document):
            continue

        assert testcase[index - 1] == node

        dom.append(dom.root(), node)

    for index, node in enumerate(rl.iter.Iterator(dom)):
        if isinstance(node, rl.Document):
            continue

        assert testcase[index - 1] == node


def test_ancestors():
    dom = rl.TreeDom()

    testcase = [
        rl.Doctype(dom, "one", "", ""),
        rl.Element(dom, "html", [(rl.QualName("lang", "", None), "en")], False, False),
        rl.Element(dom, "head", [(rl.QualName("lang", "", None), "en")], False, False),
        rl.Text(dom, "Hello World"),
    ]

    dom.append(dom.root(), testcase[0])
    dom.append(dom.root(), testcase[1])
    dom.append(testcase[1], testcase[2])
    dom.append(testcase[2], testcase[3])

    result = list(rl.iter.Ancestors(testcase[3]))

    assert isinstance(result[-1], rl.Document)
    assert result[-2] == testcase[1]


def test_children():
    dom = rl.TreeDom()

    testcase = [
        rl.Doctype(dom, "one", "", ""),
        rl.Element(dom, "html", [(rl.QualName("lang", "", None), "en")], False, False),
        rl.Element(dom, "head", [(rl.QualName("lang", "", None), "en")], False, False),
        rl.Text(dom, "Hello World"),
    ]

    for n in testcase:
        dom.append(dom.root(), n)

    result = list(rl.iter.Children(dom.root()))

    assert result == testcase


def test_traverse():
    dom = rl.TreeDom()

    testcase = [
        rl.Doctype(dom, "one", "", ""),
        rl.Element(dom, "html", [(rl.QualName("lang", "", None), "en")], False, False),
        rl.Element(dom, "head", [(rl.QualName("lang", "", None), "en")], False, False),
        rl.Text(dom, "Hello World"),
    ]

    dom.append(dom.root(), testcase[0])
    dom.append(dom.root(), testcase[1])
    dom.append(testcase[1], testcase[2])
    dom.append(testcase[2], testcase[3])

    expected = [
        (testcase[0], False),
        (testcase[0], True),
        (testcase[1], False),
        (testcase[2], False),
        (testcase[3], False),
        (testcase[3], True),
        (testcase[2], True),
        (testcase[1], True),
    ]
    result = list(rl.iter.Traverse(dom.root()))[1:-1]

    assert result == expected


def _get_text(node) -> str:
    s = ""

    for n in rl.iter.Descendants(node):
        if isinstance(n, rl.Text):
            s += n.content

    return s.strip()


def test_select():
    p = rl.Parser(rl.HtmlOptions())
    p.process(
        """<!DOCTYPE html>
<html>
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Document</title>
</head>
<body>
    <div id="header" data-role="banner">
        <h1 title="main heading">Welcome</h1>
        <p>This is paragraph 1</p>
        <p>This is paragraph 2</p>
    </div>
    <ul class="menu">
        <li class="example">Item 1</li>
        <li>Item 2</li>
        <li>Item 3</li>
    </ul>
    <a href="https://example.com" class="external">External Link</a>
    <input type="text" required>
    <input type="checkbox" checked>
    <div class="box"></div>
    <p lang="en">English</p>
    <p lang="fr">Texte en</p>
</body>
</html>"""
    )
    p.finish()
    d = p.into_dom()

    is_ok = False
    for node in rl.Select(d.root(), '[href*="example"]'):
        assert node.name == "a"

        _, v = _get_attr(node.attrs, "href")
        assert "example" in v
        is_ok = True

    assert is_ok

    is_ok = False
    for node in rl.Select(d.root(), "div p:nth-of-type(2)"):
        assert node.name == "p"

        assert _get_text(node) == "This is paragraph 2"
        is_ok = True

    assert is_ok

    is_ok = False
    for node in rl.Select(d.root(), "div:empty"):
        assert node.name == "div"

        assert node.class_list() == ["box"]

        is_ok = True

    assert is_ok

    is_ok = False
    for node in rl.Select(d.root(), "div[data-role] p"):
        assert node.name == "p"
        assert _get_text(node).startswith("This is paragraph")

        is_ok = True

    assert is_ok


def test_serialize():
    parser = rl.Parser(rl.HtmlOptions(full_document=True))
    for c in ("<html>", b"<body>Ali</body>", b"</html>"):
        parser.process(c)
    parser.finish()

    dom = parser.into_dom()

    assert rl.serialize(dom.root()) == b"<html><head></head><body>Ali</body></html>"
    assert (
        rl.serialize(dom.root(), False)
        == b'<html xmlns="http://www.w3.org/1999/xhtml"><head></head><body>Ali</body></html>'
    )

    parser = rl.Parser(rl.HtmlOptions(full_document=False))
    for c in ("<hello>", b"Ali", b"</hello>"):
        parser.process(c)
    parser.finish()

    dom = parser.into_dom()

    assert rl.serialize(dom.root(), True) == b"<html><hello>Ali</hello></html>"
    assert (
        rl.serialize(dom.root(), False)
        == b'<html xmlns="http://www.w3.org/1999/xhtml"><hello>Ali</hello></html>'
    )

    parser = rl.Parser(rl.XmlOptions())
    for c in ("<hello>", b"Ali", b"</hello>"):
        parser.process(c)
    parser.finish()

    dom = parser.into_dom()

    assert rl.serialize(dom.root()) == b"<hello>Ali</hello>"
    assert rl.serialize(dom.root(), False) == b"<hello>Ali</hello>"



import markupever
import pytest


def test_parser():  # this is a copy of test_rustlib.test_parser for markupever.parser.Parser
    parser = markupever.Parser(markupever.HtmlOptions())
    parser.process(b"<html><p>Ali</p></html>")
    parser.finish()

    repr(parser)

    parser = markupever.Parser(markupever.HtmlOptions())
    parser.process("<html><p>Ali</p></html>")

    with pytest.raises(TypeError):
        parser.process(1)

    with pytest.raises(RuntimeError):
        parser.into_dom()

    parser.finish()

    with pytest.raises(RuntimeError):
        parser.process("")

    with pytest.raises(RuntimeError):
        parser.finish()

    parser.into_dom()
    with pytest.raises(RuntimeError):
        parser.into_dom()

    parser = markupever.Parser(markupever.XmlOptions())
    for c in ("<html>", b"Ali", b"</html>"):
        parser.process(c)
    parser.finish()

    assert parser.is_finished
    assert isinstance(parser.errors(), list)
    assert parser.lineno == 1
    assert parser.quirks_mode == 2

    parser = markupever.Parser(markupever.HtmlOptions(full_document=False))
    for c in (b"<html><p>Ali</p>", "\n", "</html>"):
        parser.process(c)
    parser.finish()

    assert parser.lineno == 2

    _ = parser.into_dom()

    assert parser.is_converted

    with pytest.raises(RuntimeError):
        parser.errors()


def test_parse_function():
    assert isinstance(markupever.parse("<html></html>", markupever.XmlOptions()), markupever.dom.TreeDom)


def test_parse_file_function(tmp_path):
    import io

    file = io.BytesIO(b"<body></body>")
    assert isinstance(markupever.parse_file(file, markupever.XmlOptions()), markupever.dom.TreeDom)
    assert not file.closed
    file.close()

    file = io.StringIO("<body></body>")
    assert isinstance(markupever.parse_file(file, markupever.XmlOptions()), markupever.dom.TreeDom)
    assert not file.closed
    file.close()

    file = tmp_path / "file.html"

    with pytest.raises(FileNotFoundError):
        markupever.parse_file(str(file), markupever.HtmlOptions())

    file.write_bytes(b"<body></body>")
    assert isinstance(markupever.parse_file(str(file), markupever.HtmlOptions()), markupever.dom.TreeDom)

    markupever.parse_file(file, markupever.HtmlOptions())



import typing

QUIRKS_MODE_FULL: typing.Literal[0]
QUIRKS_MODE_LIMITED: typing.Literal[1]
QUIRKS_MODE_OFF: typing.Literal[2]

class HtmlOptions:
    """
    These are options for HTML parsing.

    Note: this type is immutable.
    """

    def __new__(
        cls: typing.Type,
        full_document=...,
        exact_errors=...,
        discard_bom=...,
        profile=...,
        iframe_srcdoc=...,
        drop_doctype=...,
        quirks_mode=...,
    ) -> "HtmlOptions":
        """
        Creates a new `HtmlOptions`

        - `full_document`: Is this a complete document? (means includes html, head, and body tag). Default: true.
        - `exact_errors`: Report all parse errors described in the spec, at some performance penalty? Default: false.
        - `discard_bom`: Discard a `U+FEFF BYTE ORDER MARK` if we see one at the beginning of the stream? Default: true.
        - `profile`: Keep a record of how long we spent in each state? Printed when `finish()` is called. Default: false.
        - `iframe_srcdoc`: Is this an `iframe srcdoc` document? Default: false.
        - `drop_doctype`: Should we drop the DOCTYPE (if any) from the tree? Default: false.
        - `quirks_mode`: Initial TreeBuilder quirks mode. Default: QUIRKS_MODE_OFF.
        """
        ...

    @property
    def full_document(self) -> bool: ...
    @property
    def exact_errors(self) -> bool: ...
    @property
    def discard_bom(self) -> bool: ...
    @property
    def profile(self) -> bool: ...
    @property
    def iframe_srcdoc(self) -> bool: ...
    @property
    def drop_doctype(self) -> bool: ...
    @property
    def quirks_mode(self) -> int: ...
    def __repr__(self) -> str: ...

class XmlOptions:
    """
    These are options for XML parsing.

    Note: this type is immutable.
    """

    def __new__(
        cls: typing.Type,
        exact_errors=...,
        discard_bom=...,
        profile=...,
    ) -> "XmlOptions":
        """
        Creates a new `XmlOptions`

        - `exact_errors`: Report all parse errors described in the spec, at some performance penalty? Default: false.
        - `discard_bom`: Discard a `U+FEFF BYTE ORDER MARK` if we see one at the beginning of the stream? Default: true.
        - `profile`: Keep a record of how long we spent in each state? Printed when `finish()` is called. Default: false.
        """
        ...

    @property
    def exact_errors(self) -> bool: ...
    @property
    def discard_bom(self) -> bool: ...
    @property
    def profile(self) -> bool: ...
    def __repr__(self) -> str: ...

class QualName:
    """
    A fully qualified name (with a namespace), used to depict names of tags and attributes.

    Namespaces can be used to differentiate between similar XML fragments. For example:

    ```
    // HTML
    <table>
      <tr>
        <td>Apples</td>
        <td>Bananas</td>
      </tr>
    </table>

    // Furniture XML
    <table>
      <name>African Coffee Table</name>
      <width>80</width>
      <length>120</length>
    </table>
    ```

    Without XML namespaces, we can't use those two fragments in the same document
    at the same time. However if we declare a namespace we could instead say:

    ```
    // Furniture XML
    <furn:table xmlns:furn="https://furniture.rs">
      <furn:name>African Coffee Table</furn:name>
      <furn:width>80</furn:width>
      <furn:length>120</furn:length>
    </furn:table>
    ```

    and bind the prefix `furn` to a different namespace.

    For this reason we parse names that contain a colon in the following way:

    ```
    <furn:table>
       |    |
       |    +- local name
       |
     prefix (when resolved gives namespace_url `https://furniture.rs`)
    ```

    Note: This type is immutable.
    """

    def __new__(
        cls,
        local: str,
        ns: typing.Union[
            str, typing.Literal["html", "xml", "xhtml", "xmlns", "xlink", "svg", "mathml", "*"]
        ] = ...,
        prefix: typing.Optional[str] = ...,
    ): ...
    @property
    def local(self) -> str:
        """The local name (e.g. `table` in `<furn:table>` above)."""
        ...
    @property
    def ns(self) -> str:
        """The namespace after resolution (e.g. https://furniture.rs in example above)."""
        ...
    @property
    def prefix(self) -> typing.Optional[str]:
        """
        The prefix of qualified (e.g. furn in <furn:table> above).
        Optional (since some namespaces can be empty or inferred),
        and only useful for namespace resolution (since different prefix can still resolve to same namespace)
        """
        ...

    def copy(self) -> "QualName": ...
    def __eq__(self, value) -> bool: ...
    def __ne__(self, value) -> bool: ...
    def __gt__(self, value) -> bool: ...
    def __ge__(self, value) -> bool: ...
    def __lt__(self, value) -> bool: ...
    def __le__(self, value) -> bool: ...
    def __hash__(self) -> int: ...
    def __repr__(self) -> str: ...

class AttrsListItems:
    def __iter__(self) -> "AttrsListItems": ...
    def __next__(self) -> typing.Tuple[QualName, str]: ...



from . import _rustlib, iterators
from ._rustlib import QualName as QualName
import typing
import itertools


class TreeDom:
    __slots__ = ("_raw",)

    def __init__(self, *, raw: typing.Optional[_rustlib.TreeDom] = None):
        """
        A tree structure which specialy designed for HTML and XML documents. Uses Rust's `Vec` type in backend.

        The memory consumed by the `TreeDom` is dynamic and depends on the number of tokens stored in the tree.
        The allocated memory is never reduced and is only released when it is dropped.
        """
        if raw is None:
            self._raw = _rustlib.TreeDom()
        else:
            assert isinstance(raw, _rustlib.TreeDom)
            self._raw = raw

    def namespaces(self) -> typing.Dict[str, str]:
        """Returns the DOM namespaces."""
        return self._raw.namespaces()

    def root(self) -> "Document":
        """Returns the root node."""
        return Document(self._raw.root())

    def select(self, expr: str, limit: int = 0, offset: int = 0) -> iterators.Select:
        """Shorthand for `self.root().select(expr, limit, offset)`"""
        return self.root().select(expr, limit, offset)

    def select_one(self, expr: str, offset: int = 0) -> typing.Optional["Element"]:
        """Shorthand for `self.root().select_one(expr, offset)`"""
        return self.root().select_one(expr, offset)

    def serialize_bytes(self, is_html: typing.Optional[bool] = None) -> bytes:
        """Shorthand for `self.root().serialize_bytes(is_html)`"""
        return self.root().serialize_bytes(is_html=is_html)  # pragma: no cover

    def serialize(self, is_html: typing.Optional[bool] = None) -> str:
        """Shorthand for `self.root().serialize(is_html)`"""
        return self.root().serialize(is_html=is_html)  # pragma: no cover

    def __iter__(self) -> typing.Generator["BaseNode", typing.Any, None]:
        """Iterates the nodes in insert order - don't matter which are orphan which not."""
        for rn in _rustlib.iter.Iterator(self._raw):
            yield BaseNode._wrap(rn)

    def __eq__(self, val: "TreeDom") -> bool:
        if not isinstance(val, TreeDom):
            return False

        return self._raw == val._raw

    def __len__(self) -> int:
        """Returns the number of nodes in tree."""
        return len(self._raw)

    def __str__(self) -> str:
        return str(self._raw)

    def __repr__(self) -> str:
        return repr(self._raw)


class _ConfigNode:
    __slots__ = ("basetype", "invalid_ordering")

    def __init__(self, basetype: typing.Optional[type], invalid_ordering: typing.Tuple[int]):
        self.basetype = basetype
        self.invalid_ordering = invalid_ordering


class Ordering:
    """Enum for `create_*` methods of a node."""

    APPEND = 0
    """Means create and append the node as the `last_child`."""
    PREPEND = 1
    """Means create and append the node as the `first_child`."""
    AFTER = 2
    """Means create and insert the node as the `next_sibling`."""
    BEFORE = 3
    """Means create and insert the node as the `prev_sibling`."""


class BaseNode:
    __slots__ = ("_raw",)

    _CONFIG: _ConfigNode = _ConfigNode(None, ())
    _SUBCLASS_WRAP = {}

    def __init__(self, node: typing.Any):
        if self._CONFIG.basetype is not None and not isinstance(node, self._CONFIG.basetype):
            raise TypeError(
                "expected {} for node, got {} - It's recommended to use nodes `create_*` methods for creating nodes and don't call directly markupever.nodes classes.".format(
                    self._CONFIG.basetype.__name__, type(node).__name__
                )
            )

        if not _rustlib._is_node_impl(node):
            raise TypeError(
                "expected one of _rustlib nodes implementations (such as _rustlib.Element, _rustlib.Comment, ...), got {}".format(
                    type(node).__name__
                )
            )

        self._raw = node

    @classmethod
    def _wrap(cls, node: typing.Any) -> "BaseNode":
        try:
            _type = cls._SUBCLASS_WRAP[type(node)]
        except KeyError:
            raise TypeError(
                "the type of node is not acceptable ({}).".format(type(node).__name__)
            ) from None

        return _type(node)

    def _connect_node(self, ordering: int, dom, child):
        if ordering in self._CONFIG.invalid_ordering:
            raise ValueError("This ordering value is not acceptable for this type.")

        if ordering == Ordering.APPEND:
            dom.append(self._raw, child)

        elif ordering == Ordering.PREPEND:
            dom.prepend(self._raw, child)

        elif ordering == Ordering.AFTER:
            dom.insert_after(self._raw, child)

        elif ordering == Ordering.BEFORE:
            dom.insert_before(self._raw, child)

        else:
            raise ValueError(
                "ordering must be one of Ordering variables like Ordering.APPEND, Ordering.PREPEND, ..."
            )

    def __init_subclass__(cls):
        assert cls._CONFIG.basetype is not None
        BaseNode._SUBCLASS_WRAP[cls._CONFIG.basetype] = cls

    @property
    def parent(self) -> typing.Optional["BaseNode"]:
        """Returns the parent of this node."""
        parent = self._raw.parent()
        return BaseNode._wrap(parent) if parent is not None else None

    @property
    def prev_sibling(self) -> typing.Optional["BaseNode"]:
        """Returns the previous sibling of this node."""
        prev_sibling = self._raw.prev_sibling()
        return BaseNode._wrap(prev_sibling) if prev_sibling is not None else None

    @property
    def next_sibling(self) -> typing.Optional["BaseNode"]:
        """Returns the next sibling of this node."""
        next_sibling = self._raw.next_sibling()
        return BaseNode._wrap(next_sibling) if next_sibling is not None else None

    @property
    def first_child(self) -> typing.Optional["BaseNode"]:
        """Returns the first child of this node."""
        first_child = self._raw.first_child()
        return BaseNode._wrap(first_child) if first_child is not None else None

    @property
    def last_child(self) -> typing.Optional["BaseNode"]:
        """Returns the last child of this node."""
        last_child = self._raw.last_child()
        return BaseNode._wrap(last_child) if last_child is not None else None

    @property
    def has_siblings(self) -> bool:
        """Returns `True` if the node has sibling."""
        return self._raw.has_siblings

    @property
    def has_children(self) -> bool:
        """Returns `True` if the node has children."""
        return self._raw.has_children

    def tree(self) -> "TreeDom":
        """Returns the tree which this node connected to."""
        return TreeDom(raw=self._raw.tree())

    def children(self) -> iterators.Children:
        """Returns an iterator which iterates over children of this node."""
        return iterators.Children(self)

    def ancestors(self) -> iterators.Ancestors:
        """Returns an iterator which iterates over ancestors (parents) of this node."""
        return iterators.Ancestors(self)

    def prev_siblings(self) -> iterators.PrevSiblings:
        """Returns an iterator which iterates over previous siblings of this node."""
        return iterators.PrevSiblings(self)

    def next_siblings(self) -> iterators.NextSiblings:
        """Returns an iterator which iterates over next siblings of this node."""
        return iterators.NextSiblings(self)

    def first_children(self) -> iterators.FirstChildren:
        """Returns an iterator which iterates over first children."""
        return iterators.FirstChildren(self)  # pragma: no cover

    def last_children(self) -> iterators.LastChildren:
        """Returns an iterator which iterates over last children."""
        return iterators.LastChildren(self)  # pragma: no cover

    def traverse(self) -> iterators.Traverse:
        """Returns a traverse iterator."""
        return iterators.Traverse(self)

    def descendants(self) -> iterators.Descendants:
        """Returns an iterator which iterates over this node and its descendants."""
        return iterators.Descendants(self)

    def attach(self, node: "BaseNode", *, ordering: int = Ordering.APPEND) -> None:
        """
        Attaches and connect `node` to this node depends on `ordering` value.

        This is not important the `node` has detached or not.
        """
        if isinstance(node._raw, _rustlib.Document):
            raise ValueError("you cannot attach a Document node to another node.")

        self._connect_node(ordering, self._raw.tree(), node._raw)

    def detach(self) -> None:
        """
        Detaches this node from other nodes (means makes it an orphan node).

        Note: you cannot detach a node move it to another tree.
        """
        if isinstance(self._raw, _rustlib.Document):
            raise ValueError("you cannot detach Document node.")

        self._raw.tree().detach(self._raw)

    def select(self, expr: str, limit: int = 0, offset: int = 0) -> iterators.Select:
        """
        Returns an iterator that uses CSS selectors to match and find nodes.

        You can use a group of CSS selectors (seperated by comma).
        """
        return iterators.Select(self, expr, limit=limit, offset=offset)

    def select_one(self, expr: str, offset: int = 0) -> typing.Optional["Element"]:
        """
        Works like `self.select(expr, offset=offset)` but only returns the first match.
        """
        selector = iterators.Select(self, expr, limit=1, offset=offset)

        try:
            node = next(selector)
        except StopIteration:
            return None
        else:
            return node

    def strings(self, strip: bool = False):
        """Go through this node descendants and yields texts."""
        for descendant in self.descendants():
            if not isinstance(descendant, Text):
                continue

            if strip:
                yield descendant.content.strip()
            else:
                yield descendant.content

    def text(self, seperator: str = "", strip: bool = False) -> str:
        """Returns the text of this node."""
        return seperator.join(self.strings(strip=strip))

    def serialize_bytes(self, is_html: typing.Optional[bool] = None) -> bytes:
        """Serialize the tree (starts from this node) to bytes."""
        return _rustlib.serialize(self._raw, is_html)

    def serialize(self, is_html: typing.Optional[bool] = None) -> str:
        """Serialize the tree (starts from this node) to string."""
        return self.serialize_bytes(is_html).decode("utf-8")

    def __eq__(self, value):
        if isinstance(value, BaseNode):
            value = value._raw

        return self._raw == value

    def __ne__(self, value):  # pragma: no cover
        if isinstance(value, BaseNode):
            value = value._raw

        return self._raw != value

    def __le__(self, value):  # pragma: no cover
        if isinstance(value, BaseNode):
            value = value._raw

        return self._raw <= value

    def __lt__(self, value):  # pragma: no cover
        if isinstance(value, BaseNode):
            value = value._raw

        return self._raw < value

    def __ge__(self, value):  # pragma: no cover
        if isinstance(value, BaseNode):
            value = value._raw

        return self._raw >= value

    def __gt__(self, value):  # pragma: no cover
        if isinstance(value, BaseNode):
            value = value._raw

        return self._raw > value

    def __repr__(self) -> str:  # pragma: no cover
        return repr(self._raw)


class Document(BaseNode):
    """The root of a document."""

    _CONFIG = _ConfigNode(_rustlib.Document, (Ordering.AFTER, Ordering.BEFORE))

    def create_doctype(
        self,
        name: str,
        public_id: str = "",
        system_id: str = "",
        *,
        ordering: int = Ordering.APPEND,
    ) -> "Doctype":
        """
        Create and connect a `Doctype` to this node depends on `ordering` value.
        """
        dom = self._raw.tree()
        node = _rustlib.Doctype(dom, name, public_id, system_id)
        self._connect_node(ordering, dom, node)
        return Doctype(node)

    def create_comment(self, content: str, *, ordering: int = Ordering.APPEND) -> "Comment":
        """
        Create and connect a `Comment` to this node depends on `ordering` value.
        """
        dom = self._raw.tree()
        node = _rustlib.Comment(dom, content)
        self._connect_node(ordering, dom, node)
        return Comment(node)

    def create_text(self, content: str, *, ordering: int = Ordering.APPEND) -> "Text":
        """
        Create and connect a `Text` to this node depends on `ordering` value.
        """
        dom = self._raw.tree()
        node = _rustlib.Text(dom, content)
        self._connect_node(ordering, dom, node)
        return Text(node)

    def create_element(
        self,
        name: str,
        attrs: typing.Union[
            typing.Sequence[typing.Tuple[typing.Union[_rustlib.QualName, str], str]],
            typing.Dict[typing.Union[_rustlib.QualName, str], str],
        ] = [],
        template: bool = False,
        mathml_annotation_xml_integration_point: bool = False,
        *,
        ordering: int = Ordering.APPEND,
    ) -> "Element":
        """
        Create and connect a `Element` to this node depends on `ordering` value.
        """
        dom = self._raw.tree()

        if isinstance(attrs, dict):
            attrs = list(attrs.items())

        node = _rustlib.Element(dom, name, attrs, template, mathml_annotation_xml_integration_point)
        self._connect_node(ordering, dom, node)
        return Element(node)

    def create_processing_instruction(
        self, data: str, target: str, *, ordering: int = Ordering.APPEND
    ) -> "ProcessingInstruction":
        """
        Create and connect a `ProcessingInstruction` to this node depends on `ordering` value.
        """
        dom = self._raw.tree()
        node = _rustlib.ProcessingInstruction(dom, data, target)
        self._connect_node(ordering, dom, node)
        return ProcessingInstruction(node)


class Doctype(BaseNode):
    """
    the doctype is the required <!doctype html> preamble found at the top of all documents.
    Its sole purpose is to prevent a browser from switching into so-called "quirks mode"
    when rendering a document; that is, the <!doctype html> doctype ensures that the browser makes
    a best-effort attempt at following the relevant specifications, rather than using a different
    rendering mode that is incompatible with some specifications.
    """

    _CONFIG = _ConfigNode(_rustlib.Doctype, (Ordering.APPEND, Ordering.PREPEND))

    @property
    def name(self) -> str:
        return self._raw.name

    @name.setter
    def name(self, value: str) -> None:
        self._raw.name = value

    @property
    def system_id(self) -> str:
        return self._raw.system_id

    @system_id.setter
    def system_id(self, value: str) -> None:
        self._raw.system_id = value

    @property
    def public_id(self) -> str:
        return self._raw.public_id

    @public_id.setter
    def public_id(self, value: str) -> None:
        self._raw.public_id = value


class Comment(BaseNode):
    """
    The Comment interface represents textual notations within markup; although it is generally not
    visually shown, such comments are available to be read in the source view.

    Comments are represented in HTML and XML as content between <!-- and -->. In XML,
    like inside SVG or MathML markup, the character sequence -- cannot be used within a comment.
    """

    _CONFIG = _ConfigNode(_rustlib.Comment, (Ordering.APPEND, Ordering.PREPEND))

    @property
    def content(self) -> str:
        return self._raw.content

    @content.setter
    def content(self, value: str) -> None:
        self._raw.content = value

    def __eq__(self, value):
        if isinstance(value, str):
            return self._raw.content == value

        return super().__eq__(value)

    def __ne__(self, value):  # pragma: no cover
        if isinstance(value, str):
            return self._raw.content != value

        return super().__ne__(value)

    def __le__(self, value):  # pragma: no cover
        if isinstance(value, str):
            return self._raw.content <= value

        return super().__le__(value)

    def __lt__(self, value):  # pragma: no cover
        if isinstance(value, str):
            return self._raw.content < value

        return super().__lt__(value)

    def __ge__(self, value):  # pragma: no cover
        if isinstance(value, str):
            return self._raw.content >= value

        return super().__ge__(value)

    def __gt__(self, value):  # pragma: no cover
        if isinstance(value, str):
            return self._raw.content > value

        return super().__gt__(value)


class Text(BaseNode):
    """A text node."""

    _CONFIG = _ConfigNode(_rustlib.Text, (Ordering.APPEND, Ordering.PREPEND))

    @property
    def content(self) -> str:
        return self._raw.content

    @content.setter
    def content(self, value: str) -> None:
        self._raw.content = value

    def __eq__(self, value):
        if isinstance(value, str):
            return self._raw.content == value

        return super().__eq__(value)

    def __ne__(self, value):  # pragma: no cover
        if isinstance(value, str):
            return self._raw.content != value

        return super().__ne__(value)

    def __le__(self, value):  # pragma: no cover
        if isinstance(value, str):
            return self._raw.content <= value

        return super().__le__(value)

    def __lt__(self, value):  # pragma: no cover
        if isinstance(value, str):
            return self._raw.content < value

        return super().__lt__(value)

    def __ge__(self, value):  # pragma: no cover
        if isinstance(value, str):
            return self._raw.content >= value

        return super().__ge__(value)

    def __gt__(self, value):  # pragma: no cover
        if isinstance(value, str):
            return self._raw.content > value

        return super().__gt__(value)


_D = typing.TypeVar("_D")


class AttrsList:
    """
    This type is only designed for communicating with element attributes. 
    
    Really it's a list, but has a behaviour between dictionary and list to provide you easy-to-use management.
    """
    __slots__ = ("__raw",)

    def __init__(self, attrs: _rustlib.AttrsList):
        self.__raw = attrs

    def append(self, key: typing.Union[_rustlib.QualName, str], value: str):
        """
        Appends a key-value pair into attributes list.
        """
        self.__raw.push(key, value)

    def insert(self, index: int, key: typing.Union[_rustlib.QualName, str], value: str):
        """
        Inserts a key-value pair at position `index` within the list, shifting all elements after it to the right.
        """
        self.__raw.insert(index, key, value)

    def _find_by_key(
        self,
        key: typing.Union[_rustlib.QualName, str],
        default: _D = None,
        start: int = 0,
    ) -> typing.Tuple[typing.Union[str, _D], int]:
        for index, item in itertools.islice(enumerate(self.__raw.items()), start, None):
            k, v = item

            if k == key:
                return v, index

        return default, -1

    def _find_by_item(
        self,
        key: typing.Union[_rustlib.QualName, str],
        value: str,
        start: int = 0,
    ) -> int:
        for index, item in itertools.islice(enumerate(self.__raw.items()), start, None):
            k, v = item

            if k == key and value == v:
                return index

        return -1

    def index(
        self, key: typing.Union[typing.Union[_rustlib.QualName, str], tuple], start: int = 0
    ) -> int:
        """
        Returns the first match index.
        """
        if isinstance(key, tuple):
            index = self._find_by_item(*key, start=start)
        else:
            _, index = self._find_by_key(key, start=start)

        if index == -1:
            raise ValueError(key)

        return index

    def get(
        self, key: typing.Union[_rustlib.QualName, str], default: _D = None, start: int = 0
    ) -> typing.Union[str, _D]:
        """
        Return the value for key if key is in the dictionary, else default.
        """
        val, index = self._find_by_key(key, start=start)
        if index == -1:
            return default

        return val

    def dedup(self) -> None:
        """
        Removes consecutive repeated elements in the attributes list.
        """
        self.__raw.dedup()  # pragma: no cover

    def pop(self, index: int = -1) -> typing.Tuple[_rustlib.QualName, str]:
        """
        Remove and return item at index (default last).

        Raises `IndexError` if list is empty or index is out of range.
        """
        if index < 0:
            index = len(self.__raw) + index

        return self.__raw.remove(index)

    def remove(
        self, key: typing.Union[typing.Union[_rustlib.QualName, str], tuple], start: int = 0
    ) -> None:
        """
        Remove first occurrence of value.

        Raises ValueError if the value is not present.
        """
        index = self.index(key, start=start)
        self.__raw.remove(index)

    def reverse(self) -> None:
        """Reverses the order of elements in the list."""
        self.__raw.reverse()

    def extend(self, m: typing.Union[dict, typing.Iterable[tuple]]):
        """
        Extend the attributes list by appending key-value pairs from the iterable or dictionary.
        """
        if isinstance(m, dict):
            m = m.items()

        for key, val in m:
            self.__raw.push(key, val)

    def clear(self) -> None:
        """Clears the attributes list, removing all values."""
        self.__raw.clear()

    def keys(self) -> typing.Generator[QualName, None, None]:
        """Returns a generator of attribute keys."""
        return (i for i, _ in self.__raw.items())

    def values(self) -> typing.Generator[str, None, None]:
        """Returns a generator of attribute values."""
        return (i for _, i in self.__raw.items())

    def __len__(self) -> int:
        """Returns `len(self)`."""
        return len(self.__raw)

    def __iter__(self):
        """Returns a generator of attribute keys."""
        return self.keys()

    def __contains__(self, key: typing.Union[typing.Union[_rustlib.QualName, str], tuple]) -> bool:
        """
        Returns `True` if the list has the specified key, else `False`.
        """
        if isinstance(key, tuple):
            index = self._find_by_item(*key)
        else:
            _, index = self._find_by_key(key)

        return index > -1

    def __delitem__(self, index: typing.Union[int, str, _rustlib.QualName]) -> None:
        """
        - If `index` is `str` or `QualName`: Finds and removes the first match from list.
        - If `index` is `int`: Removes the `index` from list.
        """
        if not isinstance(index, int):
            _, index = self._find_by_key(index)
            if index == -1:
                raise KeyError(index)

        self.__raw.remove(index)

    def __setitem__(
        self,
        index: typing.Union[int, str, _rustlib.QualName],
        val: typing.Union[str, typing.Tuple[typing.Union[_rustlib.QualName, str], str]],
    ) -> None:
        if not isinstance(index, int):
            if isinstance(val, str):
                val = (index, val)

            _, index = self._find_by_key(index)
            if index == -1:
                self.__raw.push(*val)
                return

        self.__raw.update_item(index, val[0], val[1])

    @typing.overload
    def __getitem__(
        self, index: typing.Union[str, _rustlib.QualName]
    ) -> str: ...

    @typing.overload
    def __getitem__(
        self, index: int
    ) -> typing.Tuple[_rustlib.QualName, str]: ...

    def __getitem__(self, index):
        if not isinstance(index, int):
            _, index = self._find_by_key(index)
            if index == -1:
                raise KeyError(index)

            _, val = self.__raw.get_by_index(index)
            return val

        return self.__raw.get_by_index(index)

    def __repr__(self) -> str:
        return repr(self.__raw)


class Element(BaseNode):
    """An element node."""

    _CONFIG = _ConfigNode(_rustlib.Element, ())

    @property
    def name(self) -> _rustlib.QualName:
        return self._raw.name

    @name.setter
    def name(self, value: typing.Union[str, _rustlib.QualName]) -> None:
        self._raw.name = value

    @property
    def attrs(self) -> AttrsList:
        return AttrsList(self._raw.attrs)

    @attrs.setter
    def attrs(
        self,
        value: typing.Union[
            typing.Sequence[typing.Tuple[typing.Union[_rustlib.QualName, str], str]],
            typing.Dict[typing.Union[_rustlib.QualName, str], str],
        ],
    ) -> None:
        if isinstance(value, dict):
            value = list(value.items())

        self._raw.attrs = value

    @property
    def template(self) -> bool:
        return self._raw.template

    @template.setter
    def template(self, value: bool) -> None:
        self._raw.template = value

    @property
    def mathml_annotation_xml_integration_point(self) -> bool:
        return self._raw.mathml_annotation_xml_integration_point

    @mathml_annotation_xml_integration_point.setter
    def mathml_annotation_xml_integration_point(self, value: bool) -> None:
        self._raw.mathml_annotation_xml_integration_point = value

    @property
    def id(self) -> typing.Optional[str]:
        return self._raw.id()

    @property
    def class_list(self) -> typing.List[str]:
        return self._raw.class_list()

    def create_doctype(
        self,
        name: str,
        public_id: str = "",
        system_id: str = "",
        *,
        ordering: int = Ordering.APPEND,
    ) -> "Doctype":  # pragma: no cover # it is a copy of Document.create_doctype
        """
        Create and connect a `Doctype` to this node depends on `ordering` value.
        """
        dom = self._raw.tree()
        node = _rustlib.Doctype(dom, name, public_id, system_id)
        self._connect_node(ordering, dom, node)
        return Doctype(node)

    def create_comment(
        self, content: str, *, ordering: int = Ordering.APPEND
    ) -> "Comment":  # pragma: no cover # it is a copy of Document.create_comment
        """
        Create and connect a `Comment` to this node depends on `ordering` value.
        """
        dom = self._raw.tree()
        node = _rustlib.Comment(dom, content)
        self._connect_node(ordering, dom, node)
        return Comment(node)

    def create_text(
        self, content: str, *, ordering: int = Ordering.APPEND
    ) -> "Text":  # pragma: no cover # it is a copy of Document.create_text
        """
        Create and connect a `Text` to this node depends on `ordering` value.
        """
        dom = self._raw.tree()
        node = _rustlib.Text(dom, content)
        self._connect_node(ordering, dom, node)
        return Text(node)

    def create_element(
        self,
        name: str,
        attrs: typing.Union[
            typing.Sequence[typing.Tuple[typing.Union[_rustlib.QualName, str], str]],
            typing.Dict[typing.Union[_rustlib.QualName, str], str],
        ] = [],
        template: bool = False,
        mathml_annotation_xml_integration_point: bool = False,
        *,
        ordering: int = Ordering.APPEND,
    ) -> "Element":  # pragma: no cover # it is a copy of Document.create_element
        """
        Create and connect a `Element` to this node depends on `ordering` value.
        """
        dom = self._raw.tree()

        if isinstance(attrs, dict):
            attrs = list(attrs.items())

        node = _rustlib.Element(dom, name, attrs, template, mathml_annotation_xml_integration_point)
        self._connect_node(ordering, dom, node)
        return Element(node)

    def create_processing_instruction(
        self, data: str, target: str, *, ordering: int = Ordering.APPEND
    ) -> (
        "ProcessingInstruction"
    ):  # pragma: no cover # it is a copy of Document.create_processing_instruction
        """
        Create and connect a `ProcessingInstruction` to this node depends on `ordering` value.
        """
        dom = self._raw.tree()
        node = _rustlib.ProcessingInstruction(dom, data, target)
        self._connect_node(ordering, dom, node)
        return ProcessingInstruction(node)


class ProcessingInstruction(BaseNode):
    """
    The ProcessingInstruction interface represents a processing instruction; that is,
    a Node which embeds an instruction targeting a specific application but that can
    be ignored by any other applications which don't recognize the instruction.
    """
    _CONFIG = _ConfigNode(_rustlib.ProcessingInstruction, (Ordering.APPEND, Ordering.PREPEND))

    @property
    def target(self) -> str:
        return self._raw.target

    @target.setter
    def target(self, value: str) -> None:
        self._raw.target = value

    @property
    def data(self) -> str:
        return self._raw.data

    @data.setter
    def data(self, value: str) -> None:
        self._raw.data = value



from markupever import _rustlib
import markupever
import pytest


def test_treedom():
    dom = markupever.dom.TreeDom()

    assert dom.namespaces() == {}
    assert len(dom) == 1

    _ = str(dom)
    _ = repr(dom)

    lst = list(dom)

    assert dom == dom
    assert dom != 1
    assert len(lst) == 1
    assert isinstance(lst[0], markupever.dom.Document)


def _test_rustlib_node_convert(typ, expected, dom, *args, **kwargs) -> markupever.dom.BaseNode:
    instance = markupever.dom.BaseNode._wrap(typ(dom._raw, *args, **kwargs))
    assert isinstance(instance, expected)
    return instance


def test_basenode_init():
    dom = markupever.dom.TreeDom()

    assert isinstance(dom.root(), markupever.dom.Document)

    _test_rustlib_node_convert(_rustlib.Doctype, markupever.dom.Doctype, dom, "name", "", "")
    _test_rustlib_node_convert(_rustlib.Comment, markupever.dom.Comment, dom, "content")
    _test_rustlib_node_convert(_rustlib.Text, markupever.dom.Text, dom, "content")
    _test_rustlib_node_convert(_rustlib.Element, markupever.dom.Element, dom, "name", [], False, False)
    _test_rustlib_node_convert(
        _rustlib.ProcessingInstruction, markupever.dom.ProcessingInstruction, dom, "name", "data"
    )

    with pytest.raises(TypeError):
        markupever.dom.BaseNode("invalid type")

    markupever.dom.Doctype(_rustlib.Doctype(dom._raw, "m", "", ""))

    with pytest.raises(TypeError):
        markupever.dom.Element(_rustlib.Doctype(dom._raw, "m", "", ""))

    with pytest.raises(TypeError):
        markupever.dom.BaseNode._wrap(1)


def test_connect_node():
    dom = markupever.dom.TreeDom()
    root = dom.root()

    html = root.create_element(markupever.dom.QualName("html", "html"), {"lang": "en"}, False, False)
    assert isinstance(html, markupever.dom.Element)
    assert html.parent == root
    assert html.name == "html"

    head = html.create_element("head")
    assert isinstance(head, markupever.dom.Element)
    assert head.parent == html
    assert head.name == "head"

    meta_viewport = head.create_element(
        markupever.dom.QualName("meta", "html"),
        [("name", "viewport"), ("content", "width=device-width, initial-scale=1.0")],
    )
    assert isinstance(meta_viewport, markupever.dom.Element)
    assert meta_viewport.parent == head
    assert meta_viewport.name == markupever.dom.QualName("meta", "html")

    meta_charset = meta_viewport.create_element(
        markupever.dom.QualName("meta", "html"),
        {"charset": "UTF-8"},
        ordering=markupever.dom.Ordering.BEFORE,
    )
    assert isinstance(meta_charset, markupever.dom.Element)
    assert meta_charset.parent == head
    assert meta_charset.name == "meta"
    assert meta_charset.next_sibling == meta_viewport

    body = head.create_element("body", {"class": "bg-dark"}, ordering=markupever.dom.Ordering.AFTER)
    assert isinstance(body, markupever.dom.Element)
    assert body.parent == html
    assert body.name == "body"

    with pytest.raises(ValueError):
        root.create_doctype("html", ordering=markupever.dom.Ordering.AFTER)

    with pytest.raises(ValueError):
        root.create_doctype("html", ordering=10)

    doctype = root.create_doctype("html", ordering=markupever.dom.Ordering.PREPEND)
    assert isinstance(doctype, markupever.dom.Doctype)
    assert doctype.parent == root
    assert doctype.name == "html"
    assert doctype.next_sibling == html
    assert html.prev_sibling == doctype

    doctype.name = "HTML"
    doctype.public_id = "public"
    doctype.system_id = "system"
    assert doctype.name == "HTML"
    assert doctype.public_id == "public"
    assert doctype.system_id == "system"

    assert dom.namespaces() == {"": "http://www.w3.org/1999/xhtml"}

    p = body.create_element(
        markupever.dom.QualName("p", "namespace1", "ns1"),
        {"class": "font-sans"},
        ordering=markupever.dom.Ordering.APPEND,
    )
    assert isinstance(p, markupever.dom.Element)
    assert p.parent == body
    assert p.name == "p"

    assert dom.namespaces() == {"": "http://www.w3.org/1999/xhtml", "ns1": "namespace1"}

    comment = p.create_comment("content")
    assert isinstance(comment, markupever.dom.Comment)
    assert comment.parent == p
    assert comment.content == "content"
    assert comment == "content"

    comment.content += "testme"
    assert comment == "contenttestme"

    assert p.text() == ""

    text = p.create_text("\ncontent")
    assert isinstance(text, markupever.dom.Text)
    assert text.parent == p
    assert text.content == "\ncontent"
    assert text == "\ncontent"

    text.content += " 1"

    p.create_text("\ncontent 2")

    assert p.text() == "\ncontent 1\ncontent 2"
    assert p.text(strip=True) == "content 1content 2"
    assert p.text(seperator="\t", strip=True) == "content 1\tcontent 2"

    assert text.has_siblings
    assert p.has_children
    assert p.tree() == dom

    assert (
        root.serialize()
        == '<!DOCTYPE HTML PUBLIC "public" SYSTEM "system"><html lang="en"><head><meta charset="UTF-8"><meta name="viewport" content="width=device-width, initial-scale=1.0"></head><body class="bg-dark"><p class="font-sans"><!--contenttestme-->\ncontent 1\ncontent 2</p></body></html>'
    )

    with pytest.raises(ValueError):
        root.create_comment("content", ordering=markupever.dom.Ordering.AFTER)

    with pytest.raises(ValueError):
        root.create_comment("content", ordering=markupever.dom.Ordering.BEFORE)

    comment = root.create_comment("content")
    assert isinstance(comment, markupever.dom.Comment)
    assert comment.content == "content"
    assert comment == "content"

    text = root.create_text("content")
    assert isinstance(text, markupever.dom.Text)
    assert text.content == "content"
    assert text == "content"

    pi = root.create_processing_instruction("data", "target")
    assert isinstance(pi, markupever.dom.ProcessingInstruction)
    assert pi.data == "data"
    assert pi.target == "target"

    pi.data = "md"
    pi.target = "mt"
    assert pi.data == "md"
    assert pi.target == "mt"

    assert root.first_child == doctype
    assert root.last_child == pi


def test_children():
    dom = markupever.dom.TreeDom()
    root = dom.root()

    testcases = [
        root.create_element("first"),
        root.create_text("second"),
        root.create_doctype("third"),
        root.create_processing_instruction("fourth", "target"),
        root.create_comment("fifth"),
        root.create_element("sixth"),
    ]

    for index, child in enumerate(root.children()):
        assert testcases[index] == child


def test_ancestors():
    dom = markupever.dom.TreeDom()
    root = dom.root()

    testcases = [root]
    parent = root
    for i in range(10):
        elem = parent.create_element(f"test{i}")
        testcases.append(elem)
        parent = elem

    testcases.pop()

    for index, ans in enumerate(parent.ancestors()):
        assert testcases[len(testcases) - (index + 1)] == ans


def test_siblings():
    dom = markupever.dom.TreeDom()
    root = dom.root()

    testcases = [
        root.create_element("first"),
        root.create_text("second"),
        root.create_doctype("third"),
        root.create_processing_instruction("fourth", "target"),
        root.create_comment("fifth"),
        root.create_element("sixth"),
    ]

    for index, sibling in enumerate(testcases[0].next_siblings()):
        assert testcases[index + 1] == sibling

    for index, sibling in enumerate(testcases[-1].prev_siblings()):
        assert testcases[len(testcases) - (index + 2)] == sibling

    # No need to test traverse - it's tested in Rust
    for edge in root.traverse():
        assert isinstance(edge.closed, bool)
        assert isinstance(edge.node, markupever.dom.BaseNode)
        repr(edge)


def test_detach():
    dom = markupever.dom.TreeDom()
    root = dom.root()

    with pytest.raises(ValueError):
        root.detach()

    with pytest.raises(ValueError):
        tmp_dom = markupever.dom.TreeDom()
        root.attach(tmp_dom.root())

    html = root.create_element("html")
    body = html.create_element("body")

    assert body.parent == html

    body.detach()
    assert body.parent is None

    root.attach(body)
    assert body.parent == root
    assert root.last_child == body

    html.attach(body)
    assert body.parent == html


def test_select():
    dom = markupever.parse(
        """<div class="title">
        <nav class="navbar">
            <p id="title">Hello World</p><p id="text">Hello World</p>
        </nav>
        <nav class="nav2"><p>World</p></nav>
        </div>""",
        markupever.HtmlOptions(),
    )

    count = 0
    for tag in dom.select("p:nth-child(1)"):
        count += 1
        assert isinstance(tag, markupever.dom.Element)
        assert tag.name == "p"

    assert count == 2

    count = 0
    for tag in dom.select("p:nth-child(1)", limit=1):
        count += 1
        assert isinstance(tag, markupever.dom.Element)
        assert tag.name == "p"

    assert count == 1

    count = 0
    for tag in dom.select("p"):
        count += 1
        assert isinstance(tag, markupever.dom.Element)
        assert tag.name == "p"

    assert count == 3

    count = 0
    for tag in dom.select("p", limit=1):
        count += 1
        assert isinstance(tag, markupever.dom.Element)
        assert tag.name == "p"
        assert tag.id == "title"

    assert count == 1

    count = 0
    for tag in dom.select("p", offset=3):
        count += 1
        assert isinstance(tag, markupever.dom.Element)
        assert tag.name == "p"
        assert tag.id is None

    assert count == 1

    tag = dom.select_one("nav.nav2")
    assert tag.name == "nav"
    assert tag.class_list == ["nav2"]
    assert tag.text() == "World"

    tag = dom.select_one("nav.nav2", offset=3)
    assert tag is None


def test_element():
    dom = markupever.dom.TreeDom()
    root = dom.root()

    html = root.create_element(
        "html", {"lang": "en", "class": "hello world"}, mathml_annotation_xml_integration_point=True
    )
    assert html.class_list == ["hello", "world"]
    assert html.template is False
    assert html.mathml_annotation_xml_integration_point is True

    html.name = markupever.dom.QualName("tag", "html")
    html.template = True
    html.mathml_annotation_xml_integration_point = False

    assert html.template is True
    assert html.mathml_annotation_xml_integration_point is False
    assert html.name == "tag"

    html.attrs["class"] = "hello man"
    assert html.class_list == ["hello", "man"]

    assert len(html.attrs) == 2
    del html.attrs[0]
    assert len(html.attrs) == 1

    assert html.class_list == ["hello", "man"]
    assert html.id is None

    html.attrs = [("id", "markup"), ("data-role", "button")]

    assert html.class_list == []
    html.attrs.append("class", "btn border")
    assert "btn" in html.class_list
    assert "border" in html.class_list
    assert html.attrs[0] == (markupever.dom.QualName("id"), "markup")
    assert html.id == "markup"
    assert html.attrs["id"] == "markup"
    assert html.attrs[markupever.dom.QualName("id")] == "markup"

    with pytest.raises(IndexError):
        html.attrs[10]

    with pytest.raises(KeyError):
        html.attrs["ali"]

    html.attrs.insert(0, "onclick", "alert")
    assert html.attrs[0] == (markupever.dom.QualName("onclick"), "alert")

    html.attrs = [("id", "id1"), ("id", "id2"), ("data-role", "button")]
    assert html.id == "id1"

    assert html.attrs.get("id") == "id1"
    assert html.attrs.get("id", start=1) == "id2"
    assert html.attrs.get("dt", start=1, default="h") == "h"

    assert html.attrs.index("id") == 0
    assert html.attrs.index("id", start=1) == 1
    assert html.attrs.index(("id", "id1")) == 0
    assert html.attrs.index(("id", "id2")) == 1

    with pytest.raises(ValueError):
        html.attrs.index(("id", "id3"))

    html.attrs.reverse()

    html.attrs.pop()
    assert len(html.attrs) == 2

    html.attrs.remove("data-role")
    assert len(html.attrs) == 1

    assert list(html.attrs) == [_rustlib.QualName("id")]
    assert list(html.attrs.values()) == ["id2"]

    with pytest.raises(KeyError):
        del html.attrs["a"]

    html.attrs["a"] = "b"
    del html.attrs["a"]

    html.attrs = {"a": "b"}

    assert "a" in html.attrs
    assert _rustlib.QualName("a") in html.attrs
    assert (_rustlib.QualName("a"), "b") in html.attrs

    repr(html.attrs)

    html.attrs.clear()
    assert len(html.attrs) == 0

    html.attrs.extend({"b": "c", "d": "e"})
    assert len(html.attrs) == 2



from .parser import (
    Parser as Parser,
    parse as parse,
    parse_file as parse_file,
)
from ._rustlib import (
    HtmlOptions as HtmlOptions,
    XmlOptions as XmlOptions,
    QUIRKS_MODE_FULL as QUIRKS_MODE_FULL,
    QUIRKS_MODE_OFF as QUIRKS_MODE_OFF,
    QUIRKS_MODE_LIMITED as QUIRKS_MODE_LIMITED,
)
from . import dom as dom
from . import iterators as iterators
